"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * This work is heavily based on https://github.com/raspberrypi/usbboot
 * Copyright 2016 Raspberry Pi Foundation
 */
const bluebird_1 = require("bluebird");
const _debug = require("debug");
const events_1 = require("events");
const fs_1 = require("fs");
const Path = require("path");
const usb = require("usb");
const readFile = bluebird_1.promisify(fs_1.readFile);
const debug = _debug('node-raspberrypi-usbboot');
const TRANSFER_BLOCK_SIZE = Math.pow(1024, 2);
// The equivalent of a NULL buffer, given that node-usb complains
// if the data argument is not an instance of Buffer
const NULL_BUFFER = Buffer.alloc(0);
/**
 * @summary The size of the boot message bootcode length section
 */
const BOOT_MESSAGE_BOOTCODE_LENGTH_SIZE = 4;
/**
 * @summary The offset of the boot message bootcode length section
 */
const BOOT_MESSAGE_BOOTCODE_LENGTH_OFFSET = 0;
/**
 * @summary The size of the boot message signature section
 */
const BOOT_MESSAGE_SIGNATURE_SIZE = 20;
/**
 * @summary The offset of the file message command section
 */
const FILE_MESSAGE_COMMAND_OFFSET = 0;
/**
 * @summary The size of the file message command section
 */
const FILE_MESSAGE_COMMAND_SIZE = 4;
/**
 * @summary The offset of the file message file name section
 */
const FILE_MESSAGE_FILE_NAME_OFFSET = FILE_MESSAGE_COMMAND_SIZE;
/**
 * @summary The size of the file message file name section
 */
const FILE_MESSAGE_FILE_NAME_SIZE = 256;
/**
 * @summary The GET_STATUS usb control transfer request code
 * @description
 * See http://www.jungo.com/st/support/documentation/windriver/811/wdusb_man_mhtml/node55.html#usb_standard_dev_req_codes
 */
const USB_REQUEST_CODE_GET_STATUS = 0;
/**
 * @summary The maximum buffer length of a usbboot message
 */
const USBBOOT_MESSAGE_MAX_BUFFER_LENGTH = 0xffff;
/**
 * @summary The amount of bits to shift to the right on a control transfer index
 */
const CONTROL_TRANSFER_INDEX_RIGHT_BIT_SHIFT = 16;
/**
 * @summary The size of the usbboot file message
 */
const FILE_MESSAGE_SIZE = FILE_MESSAGE_COMMAND_SIZE + FILE_MESSAGE_FILE_NAME_SIZE;
/**
 * @summary The usbboot return code that represents success
 */
const RETURN_CODE_SUCCESS = 0;
/**
 * @summary The buffer length of the return code message
 */
const RETURN_CODE_LENGTH = 4;
/**
 * @summary The timeout for USB control transfers, in milliseconds
 */
// In node-usb, 0 means "infinite" timeout
const USB_CONTROL_TRANSFER_TIMEOUT_MS = 0;
const USB_ENDPOINT_INTERFACES_SOC_BCM2835 = 1;
const USB_VENDOR_ID_BROADCOM_CORPORATION = 0x0a5c;
const USB_PRODUCT_ID_BCM2708_BOOT = 0x2763;
const USB_PRODUCT_ID_BCM2710_BOOT = 0x2764;
// When the pi reboots in mass storage mode, it has this product id
const USB_VENDOR_ID_NETCHIP_TECHNOLOGY = 0x0525;
const USB_PRODUCT_ID_POCKETBOOK_PRO_903 = 0xa4a5;
// Delay in ms after which we consider that the device was unplugged (not resetted)
const DEVICE_UNPLUG_TIMEOUT = 5000;
// Delay (in ms) to wait when epRead throws an error
const READ_ERROR_DELAY = 100;
var FileMessageCommand;
(function (FileMessageCommand) {
    FileMessageCommand[FileMessageCommand["GetFileSize"] = 0] = "GetFileSize";
    FileMessageCommand[FileMessageCommand["ReadFile"] = 1] = "ReadFile";
    FileMessageCommand[FileMessageCommand["Done"] = 2] = "Done";
})(FileMessageCommand || (FileMessageCommand = {}));
const getCommand = (fileMessageBuffer) => {
    const command = fileMessageBuffer.readInt32LE(FILE_MESSAGE_COMMAND_OFFSET);
    if (!(command in FileMessageCommand)) {
        throw new Error(`Invalid file message command code: ${command}`);
    }
    return command;
};
const getFilename = (fileMessageBuffer) => {
    // The parsed string will likely contain tons of trailing
    // null bytes that we should get rid of for convenience
    let end = fileMessageBuffer.indexOf(0, FILE_MESSAGE_FILE_NAME_OFFSET);
    if (end === -1) {
        end = fileMessageBuffer.length;
    }
    return fileMessageBuffer
        .slice(FILE_MESSAGE_FILE_NAME_OFFSET, end)
        .toString('ascii');
};
/**
 * @summary Parse a file message buffer from a device
 */
const parseFileMessageBuffer = (fileMessageBuffer) => {
    let command = getCommand(fileMessageBuffer);
    const filename = getFilename(fileMessageBuffer);
    // A blank file name can also mean "done"
    if (filename === '') {
        command = FileMessageCommand.Done;
    }
    return { command, filename };
};
/**
 * @summary Perform a USB control transfer
 * @description
 * See http://libusb.sourceforge.net/api-1.0/group__syncio.html
 */
const performControlTransfer = (device, bmRequestType, bRequest, wValue, wIndex, dataOrLength) => __awaiter(this, void 0, void 0, function* () {
    const previousTimeout = device.timeout;
    device.timeout = USB_CONTROL_TRANSFER_TIMEOUT_MS;
    const result = yield bluebird_1.fromCallback((callback) => {
        device.controlTransfer(bmRequestType, bRequest, wValue, wIndex, dataOrLength, callback);
    });
    device.timeout = previousTimeout;
    return result;
});
const isUsbBootCapableUSBDevice = (device) => {
    return (device.deviceDescriptor.idVendor === USB_VENDOR_ID_BROADCOM_CORPORATION &&
        (device.deviceDescriptor.idProduct === USB_PRODUCT_ID_BCM2708_BOOT ||
            device.deviceDescriptor.idProduct === USB_PRODUCT_ID_BCM2710_BOOT));
};
const isRaspberryPiInMassStorageMode = (device) => {
    return (device.deviceDescriptor.idVendor === USB_VENDOR_ID_NETCHIP_TECHNOLOGY &&
        device.deviceDescriptor.idProduct === USB_PRODUCT_ID_POCKETBOOK_PRO_903);
};
const initializeDevice = (device) => {
    // interface is a reserved keyword in TypeScript so we use iface
    device.open();
    // Handle 2837 where it can start with two interfaces, the first is mass storage
    // the second is the vendor interface for programming
    let interfaceNumber;
    let endpointNumber;
    if (device._configDescriptor.bNumInterfaces ===
        USB_ENDPOINT_INTERFACES_SOC_BCM2835) {
        interfaceNumber = 0;
        endpointNumber = 1;
    }
    else {
        interfaceNumber = 1;
        endpointNumber = 3;
    }
    const iface = device.interface(interfaceNumber);
    iface.claim();
    const endpoint = iface.endpoint(endpointNumber);
    debug('Initialized device correctly', portId(device));
    return { iface, endpoint };
};
const sendSize = (device, size) => __awaiter(this, void 0, void 0, function* () {
    yield performControlTransfer(device, usb.LIBUSB_REQUEST_TYPE_VENDOR, USB_REQUEST_CODE_GET_STATUS, size & USBBOOT_MESSAGE_MAX_BUFFER_LENGTH, size >> CONTROL_TRANSFER_INDEX_RIGHT_BIT_SHIFT, NULL_BUFFER);
});
function* chunks(buffer, size) {
    for (let start = 0; start < buffer.length; start += size) {
        yield buffer.slice(start, start + size);
    }
}
const epWrite = (buffer, device, endpoint) => __awaiter(this, void 0, void 0, function* () {
    yield sendSize(device, buffer.length);
    if (buffer.length > 0) {
        for (const chunk of chunks(buffer, TRANSFER_BLOCK_SIZE)) {
            yield bluebird_1.fromCallback(callback => {
                endpoint.transfer(chunk, callback);
            });
        }
    }
});
const epRead = (device, bytesToRead) => __awaiter(this, void 0, void 0, function* () {
    return yield performControlTransfer(device, usb.LIBUSB_REQUEST_TYPE_VENDOR | usb.LIBUSB_ENDPOINT_IN, USB_REQUEST_CODE_GET_STATUS, bytesToRead & USBBOOT_MESSAGE_MAX_BUFFER_LENGTH, bytesToRead >> CONTROL_TRANSFER_INDEX_RIGHT_BIT_SHIFT, bytesToRead);
});
const getDeviceId = (device) => {
    return `${device.busNumber}:${device.deviceAddress}`;
};
const safeReadFile = (filename) => __awaiter(this, void 0, void 0, function* () {
    try {
        return yield readFile(Path.join(__dirname, '..', 'blobs', filename));
    }
    catch (e) { }
});
const getFileBuffer = (filename) => __awaiter(this, void 0, void 0, function* () {
    let buffer = yield safeReadFile(filename);
    if (buffer === undefined) {
        buffer = yield safeReadFile(Path.join('raspberrypi', filename));
    }
    if (buffer === undefined) {
        debug("Can't read file", filename);
    }
    return buffer;
});
/**
 * @summary Create a boot message buffer
 *
 * @description
 * This is based on the following data structure:
 *
 * typedef struct MESSAGE_S {
 *   int length;
 *   unsigned char signature[20];
 * } boot_message_t;
 *
 * This needs to be sent to the out endpoint of the USB device
 * as a 24 bytes big-endian buffer where:
 *
 * - The first 4 bytes contain the size of the bootcode.bin buffer
 * - The remaining 20 bytes contain the boot signature, which
 *   we don't make use of in this implementation
 */
const createBootMessageBuffer = (bootCodeBufferLength) => {
    const bootMessageBufferSize = BOOT_MESSAGE_BOOTCODE_LENGTH_SIZE + BOOT_MESSAGE_SIGNATURE_SIZE;
    // Buffers are automatically filled with zero bytes
    const bootMessageBuffer = Buffer.alloc(bootMessageBufferSize);
    // The bootcode length should be stored in 4 big-endian bytes
    bootMessageBuffer.writeInt32BE(bootCodeBufferLength, BOOT_MESSAGE_BOOTCODE_LENGTH_OFFSET);
    return bootMessageBuffer;
};
const secondStageBoot = (device, endpoint) => __awaiter(this, void 0, void 0, function* () {
    const bootcodeBuffer = yield getFileBuffer('bootcode.bin');
    if (bootcodeBuffer === undefined) {
        throw new Error("Can't find bootcode.bin");
    }
    const bootMessage = createBootMessageBuffer(bootcodeBuffer.length);
    yield epWrite(bootMessage, device, endpoint);
    debug(`Writing ${bootMessage.length} bytes`, portId(device));
    yield epWrite(bootcodeBuffer, device, endpoint);
    // raspberrypi's sample code has a sleep(1) here, but it looks like it isn't required.
    const data = yield epRead(device, RETURN_CODE_LENGTH);
    const returnCode = data.readInt32LE(0);
    if (returnCode !== RETURN_CODE_SUCCESS) {
        throw new Error(`Couldn't write the bootcode, got return code ${returnCode} from device`);
    }
});
class UsbbootDevice extends events_1.EventEmitter {
    constructor(portId) {
        super();
        this.portId = portId;
        this._step = 0;
    }
    get progress() {
        return Math.round((this._step / UsbbootDevice.LAST_STEP) * 100);
    }
    get step() {
        return this._step;
    }
    set step(step) {
        this._step = step;
        this.emit('progress', this.progress);
    }
}
// LAST_STEP is hardcoded here as it is depends on the bootcode.bin file we send to the pi.
// List of steps:
// 0) device connects with iSerialNumber 0 and we write bootcode.bin to it
// 1) the device detaches
// 2 - 39) the device reattaches with iSerialNumber 1 and we upload the files it requires (the number of steps depends on the device)
// 40) the device detaches
// 41) the device reattaches as a mass storage device
UsbbootDevice.LAST_STEP = 41;
exports.UsbbootDevice = UsbbootDevice;
class UsbbootScanner extends events_1.EventEmitter {
    constructor() {
        super();
        this.usbbootDevices = new Map();
        this.boundAttachDevice = this.attachDevice.bind(this);
        this.boundDetachDevice = this.detachDevice.bind(this);
    }
    start() {
        debug('Waiting for BCM2835/6/7');
        // Prepare already connected devices
        usb.getDeviceList().map(this.boundAttachDevice);
        // At this point all devices from `usg.getDeviceList()` above
        // have had an 'attach' event emitted if they were raspberry pis.
        this.emit('ready');
        // Watch for new devices being plugged in and prepare them
        usb.on('attach', this.boundAttachDevice);
        // Watch for devices detaching
        usb.on('detach', this.boundDetachDevice);
    }
    stop() {
        usb.removeListener('attach', this.boundAttachDevice);
        usb.removeListener('detach', this.boundDetachDevice);
        this.usbbootDevices.clear();
    }
    step(device, step) {
        const usbbootDevice = this.getOrCreate(device);
        usbbootDevice.step = step;
        if (step === UsbbootDevice.LAST_STEP) {
            this.remove(device);
        }
    }
    get(device) {
        const key = portId(device);
        return this.usbbootDevices.get(key);
    }
    getOrCreate(device) {
        const key = portId(device);
        let usbbootDevice = this.usbbootDevices.get(key);
        if (usbbootDevice === undefined) {
            usbbootDevice = new UsbbootDevice(key);
            this.usbbootDevices.set(key, usbbootDevice);
            this.emit('attach', usbbootDevice);
        }
        return usbbootDevice;
    }
    remove(device) {
        const key = portId(device);
        const usbbootDevice = this.usbbootDevices.get(key);
        if (usbbootDevice !== undefined) {
            this.usbbootDevices.delete(key);
            this.emit('detach', usbbootDevice);
        }
    }
    attachDevice(device) {
        return __awaiter(this, void 0, void 0, function* () {
            if (isRaspberryPiInMassStorageMode(device) &&
                this.usbbootDevices.has(portId(device))) {
                this.step(device, 41);
                return;
            }
            if (!isUsbBootCapableUSBDevice(device)) {
                return;
            }
            debug('Found serial number', device.deviceDescriptor.iSerialNumber);
            debug('port id', portId(device));
            try {
                const { iface, endpoint } = initializeDevice(device);
                if (device.deviceDescriptor.iSerialNumber === 0) {
                    debug('Sending bootcode.bin', portId(device));
                    this.step(device, 0);
                    yield secondStageBoot(device, endpoint);
                    // The device will now detach and reattach with iSerialNumber 1.
                    // This takes approximately 1.5 seconds
                }
                else {
                    debug('Second stage boot server', portId(device));
                    yield this.fileServer(device, endpoint, 2);
                }
                device.close();
            }
            catch (error) {
                debug('error', error, portId(device));
                this.remove(device);
            }
        });
    }
    detachDevice(device) {
        if (!isUsbBootCapableUSBDevice(device)) {
            return;
        }
        const step = device.deviceDescriptor.iSerialNumber === 0 ? 1 : 40;
        debug('detach', portId(device), step);
        this.step(device, step);
        // This timeout is here to differentiate between the device resetting and the device being unplugged
        // If the step didn't changed in 5 seconds, we assume the device was unplugged.
        setTimeout(() => {
            const usbbootDevice = this.get(device);
            if (usbbootDevice !== undefined && usbbootDevice.step === step) {
                debug('device', portId(device), 'did not reattached after', DEVICE_UNPLUG_TIMEOUT, 'ms.');
                this.remove(device);
            }
        }, DEVICE_UNPLUG_TIMEOUT);
    }
    fileServer(device, endpoint, step) {
        return __awaiter(this, void 0, void 0, function* () {
            while (true) {
                this.step(device, step);
                step += 1;
                let data;
                try {
                    data = yield epRead(device, FILE_MESSAGE_SIZE);
                }
                catch (error) {
                    if (error.message === 'LIBUSB_ERROR_NO_DEVICE' ||
                        error.message === 'LIBUSB_ERROR_IO') {
                        // Drop out if the device goes away
                        break;
                    }
                    yield bluebird_1.delay(READ_ERROR_DELAY);
                    continue;
                }
                const message = parseFileMessageBuffer(data);
                debug('Received message', FileMessageCommand[message.command], message.filename, portId(device));
                if (message.command === FileMessageCommand.GetFileSize ||
                    message.command === FileMessageCommand.ReadFile) {
                    const buffer = yield getFileBuffer(message.filename);
                    if (buffer === undefined) {
                        debug(`Couldn't find ${message.filename}`, portId(device));
                        yield sendSize(device, 0);
                    }
                    else {
                        if (message.command === FileMessageCommand.GetFileSize) {
                            yield sendSize(device, buffer.length);
                        }
                        else {
                            yield epWrite(buffer, device, endpoint);
                        }
                    }
                }
                else if (message.command === FileMessageCommand.Done) {
                    break;
                }
            }
            debug('File server done', portId(device));
        });
    }
}
exports.UsbbootScanner = UsbbootScanner;
const portId = (device) => {
    return `${device.busNumber}-${device.portNumbers.join('.')}`;
};
//# sourceMappingURL=index.js.map