// Generated by CoffeeScript 1.12.7

/*
Copyright 2016 Resin.io

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */

/**
 * @module imagefs
 */
var Promise, _, checkImageType, composeDisposers, driver, filedisk, listDirectory, read, readFile, replaceStream, stream, utils, write, writeFile;

Promise = require('bluebird');

_ = require('lodash');

filedisk = require('file-disk');

replaceStream = require('replacestream');

stream = require('stream');

driver = require('./driver');

utils = require('./utils');

checkImageType = function(image) {
  if (!(_.isString(image) || image instanceof filedisk.Disk)) {
    throw new Error('image must be a String (file path) or a filedisk.Disk instance');
  }
};

composeDisposers = function(outerDisposer, createInnerDisposer) {
  return Promise.resolve(outerDisposer).then(function(outerDisposer) {
    return outerDisposer._promise.then(function(outerResult) {
      return Promise["try"](function() {
        return Promise.resolve(createInnerDisposer(outerResult)).then(function(innerDisposer) {
          return innerDisposer._promise.then(function() {
            return [innerDisposer, innerDisposer._promise];
          });
        });
      })["catch"](function(err) {
        outerDisposer._data(outerResult);
        throw err;
      }).spread(function(innerDisposer, innerResult) {
        return Promise.resolve(innerResult).disposer(function(innerResult) {
          return Promise.resolve(innerDisposer._data(innerResult)).then(function() {
            return outerDisposer._data(outerResult);
          });
        });
      });
    });
  });
};


/**
 * @summary Get a bluebird.disposer of a node fs like interface for a partition
 * @function
 * @public
 *
 * @param {String|filedisk.Disk} disk - path to the image or filedisk.Disk instance
 * @param {Number} partition - partition number
 *
 * @returns {bluebird.disposer<fs>} node fs like interface
 *
 * @example
 *
 * Promise.using imagefs.interact('/foo/bar.img', 5), (fs) ->
 *   fs.readFileAsync('/bar/qux')
 *   .then (contents) ->
 *     console.log(contents)
 */

exports.interact = function(disk, partition) {
  checkImageType(disk);
  if (_.isString(disk)) {
    return composeDisposers(filedisk.openFile(disk, 'r+'), function(fd) {
      disk = new filedisk.FileDisk(fd, true);
      return driver.interact(disk, partition);
    });
  } else if (disk instanceof filedisk.Disk) {
    return driver.interact(disk, partition);
  }
};

read = function(disk, partition, path) {
  return composeDisposers(driver.interact(disk, partition), function(fs_) {
    var outputStream, readStream, startReadStream;
    readStream = null;
    outputStream = new stream.PassThrough();
    startReadStream = function() {
      var e;
      try {
        readStream = fs_.createReadStream(path, {
          autoClose: false
        });
      } catch (error) {
        e = error;
        outputStream.emit('error', e);
        return;
      }
      readStream.on('error', function(err) {
        return outputStream.emit('error', err);
      });
      return readStream.pipe(outputStream);
    };
    outputStream.once('newListener', function() {
      return process.nextTick(startReadStream);
    });
    return Promise.resolve(outputStream).disposer(function(stream) {
      outputStream.end();
      if ((readStream != null) && readStream.closeAsync) {
        return readStream.closeAsync();
      }
    });
  });
};


/**
 * @summary Get a device file readable stream
 * @function
 * @public
 *
 * @param {Object} definition - device path definition
 * @param {String|filedisk.Disk} definition.image - path to the image or filedisk.Disk instance
 * @param {Number} [definition.partition] - partition number
 * @param {String} definition.path - file path
 *
 * @returns {bluebird.disposer<ReadStream>} file stream
 *
 * @example
 * disposer = imagefs.read
 * 	image: '/foo/bar.img'
 * 	partition: 5
 * 	path: '/baz/qux'
 *
 * Promise.using disposer, (stream) ->
 *   out = fs.createWriteStream('/bar/qux')
 *   stream.pipe(out)
 *   utils.waitStream(out)
 */

exports.read = function(definition) {
  checkImageType(definition.image);
  if (_.isString(definition.image)) {
    return composeDisposers(filedisk.openFile(definition.image, 'r'), function(fd) {
      var disk;
      disk = new filedisk.FileDisk(fd, true);
      return read(disk, definition.partition, definition.path);
    });
  } else if (definition.image instanceof filedisk.Disk) {
    return read(definition.image, definition.partition, definition.path);
  }
};

write = function(disk, partition, path, stream) {
  return Promise.using(driver.interact(disk, partition), function(fs_) {
    var outStream;
    outStream = fs_.createWriteStream(path);
    stream.pipe(outStream);
    return utils.waitStream(outStream);
  });
};


/**
 * @summary Write a stream to a device file
 * @function
 * @public
 *
 * @param {Object} definition - device path definition
 * @param {String|filedisk.Disk} definition.image - path to the image or filedisk.Disk instance
 * @param {Number} [definition.partition] - partition number
 * @param {String} definition.path - file path
 *
 * @param {ReadStream} stream - contents stream
 * @returns {Promise}
 *
 * @example
 * imagefs.write
 * 	image: '/foo/bar.img'
 * 	partition: 2
 * 	path: '/baz/qux'
 * , fs.createReadStream('/baz/qux')
 */

exports.write = function(definition, stream) {
  checkImageType(definition.image);
  if (_.isString(definition.image)) {
    return Promise.using(filedisk.openFile(definition.image, 'r+'), function(fd) {
      var disk;
      disk = new filedisk.FileDisk(fd, false, false);
      return write(disk, definition.partition, definition.path, stream);
    });
  } else if (definition.image instanceof filedisk.Disk) {
    return write(definition.image, definition.partition, definition.path, stream);
  }
};

readFile = function(disk, partition, path) {
  return Promise.using(driver.interact(disk, partition), function(fs_) {
    return fs_.readFileAsync(path, {
      encoding: 'utf8'
    });
  });
};


/**
 * @summary Read a device file
 * @function
 * @public
 *
 * @param {Object} definition - device path definition
 * @param {String|filedisk.Disk} definition.image - path to the image or filedisk.Disk instance
 * @param {Number} [definition.partition] - partition number
 * @param {String} definition.path - file path
 *
 * @returns {Promise<String>} file text
 *
 * @example
 * imagefs.readFile
 * 	image: '/foo/bar.img'
 * 	partition: 5
 * 	path: '/baz/qux'
 * .then (contents) ->
 * 	console.log(contents)
 */

exports.readFile = function(definition) {
  checkImageType(definition.image);
  if (_.isString(definition.image)) {
    return Promise.using(filedisk.openFile(definition.image, 'r'), function(fd) {
      var disk;
      disk = new filedisk.FileDisk(fd);
      return readFile(disk, definition.partition, definition.path);
    });
  } else if (definition.image instanceof filedisk.Disk) {
    return readFile(definition.image, definition.partition, definition.path);
  }
};

writeFile = function(disk, partition, path, contents) {
  return Promise.using(driver.interact(disk, partition), function(fs_) {
    return fs_.writeFileAsync(path, contents);
  });
};


/**
 * @summary Write a device file
 * @function
 * @public
 *
 * @param {Object} definition - device path definition
 * @param {String|filedisk.Disk} definition.image - path to the image or filedisk.Disk instance
 * @param {Number} [definition.partition] - partition number
 * @param {String} definition.path - file path
 *
 * @param {String} contents - contents string
 * @returns {Promise}
 *
 * @example
 * imagefs.writeFile
 * 	image: '/foo/bar.img'
 * 	partition: 2
 * 	path: '/baz/qux'
 * , 'foo bar baz'
 */

exports.writeFile = function(definition, contents) {
  checkImageType(definition.image);
  if (_.isString(definition.image)) {
    return Promise.using(filedisk.openFile(definition.image, 'r+'), function(fd) {
      var disk;
      disk = new filedisk.FileDisk(fd);
      return writeFile(disk, definition.partition, definition.path, contents);
    });
  } else if (definition.image instanceof filedisk.Disk) {
    return writeFile(definition.image, definition.partition, definition.path, contents);
  }
};


/**
 * @summary Copy a device file
 * @function
 * @public
 *
 * @param {Object} input - input device path definition
 * @param {String|filedisk.Disk} definition.image - path to the image or filedisk.Disk instance
 * @param {Number} [input.partition] - partition number
 * @param {String} input.path - file path
 *
 * @param {Object} output - output device path definition
 * @param {String} output.image - path to the image
 * @param {Number} [output.partition] - partition number
 * @param {String} output.path - file path
 *
 * @returns {Promise}
 *
 * @example
 * imagefs.copy
 * 	image: '/foo/bar.img'
 * 	partition: 2
 * 	path: '/baz/qux'
 * ,
 * 	image: '/foo/bar.img'
 * 	partition: 5
 * 	path: '/baz/hello'
 */

exports.copy = function(input, output) {
  return Promise.using(exports.read(input), function(stream) {
    return exports.write(output, stream);
  });
};


/**
 * @summary Perform search and replacement in a file
 * @function
 * @public
 *
 * @param {Object} definition - device path definition
 * @param {String|filedisk.Disk} definition.image - path to the image or filedisk.Disk instance
 * @param {Number} [definition.partition] - partition number
 * @param {String} definition.path - file path
 *
 * @param {(String|RegExp)} search - search term
 * @param {String} replace - replace value
 *
 * @returns {Promise}
 *
 * @example
 * imagefs.replace
 * 	image: '/foo/bar.img'
 * 	partition: 2
 * 	path: '/baz/qux'
 * , 'bar', 'baz'
 */

exports.replace = function(definition, search, replace) {
  return Promise.using(exports.read(definition), function(stream) {
    var replacedStream;
    replacedStream = stream.pipe(replaceStream(search, replace));
    return exports.write(definition, replacedStream);
  });
};

listDirectory = function(disk, partition, path) {
  return Promise.using(driver.interact(disk, partition), function(fs_) {
    return fs_.readdirAsync(path);
  });
};


/**
 * @summary List the contents of a directory
 * @function
 * @public
 *
 * @param {Object} definition - device path definition
 * @param {String|filedisk.Disk} definition.image - path to the image or filedisk.Disk instance
 * @param {Number} [definition.partition] - partition number
 * @param {String} definition.path - directory path
 *
 * @returns {Promise<String[]>} list of files in directory
 *
 * @example
 * imagefs.listDirectory
 * 	image: '/foo/bar.img'
 * 	partition: 5
 * 	path: '/my/directory'
 * .then (files) ->
 * 	console.log(files)
 */

exports.listDirectory = function(definition) {
  checkImageType(definition.image);
  if (_.isString(definition.image)) {
    return Promise.using(filedisk.openFile(definition.image, 'r+'), function(fd) {
      var disk;
      disk = new filedisk.FileDisk(fd);
      return listDirectory(disk, definition.partition, definition.path);
    });
  } else if (definition.image instanceof filedisk.Disk) {
    return listDirectory(definition.image, definition.partition, definition.path);
  }
};
